package client

import (
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"

	"github.com/pkg/errors"
)

const (
	storeMsgFuncName = "StoreMessageInfo"
)

type E2EEClient struct {
	gatewayClient *GatewayClient
	orgID         string
	requestIP     string
	requestPort   int
}

type encMsgForSDKParams struct {
	TargetOrgIDs   []string `json:"target_org_ids"`
	Message        string   `json:"message"`
	UpToChain      bool     `json:"up_to_chain"`
	UpdateGroupKey bool     `json:"update_group_key"`
}

type encMsgForOrgUserResponseParams struct {
	TransactionID string `json:"transaction_id"`
	DataToChain   []byte `json:"data_to_chain"`
}

type decMsgForSDKRequestParams struct {
	TransactionID string `json:"transaction_id"`
}

type decMsgForSDKResponseParams struct {
	Message string `json:"message"`
}

func NewE2EEClient(gatewayClient *GatewayClient, consensusOrgID, orgID string) (*E2EEClient, error) {
	if gatewayClient == nil {
		return nil, errors.New("gatewayClient is nil")
	}
	requestInfo, ok := gatewayClient.E2EEReqInfo[orgID]
	if !ok {
		return nil, errors.Errorf("get organization[%s] E2EE request info failed", orgID)
	}
	e2eeClient := &E2EEClient{
		gatewayClient: gatewayClient,
		orgID:         orgID,
		requestIP:     requestInfo.RequestIp,
		requestPort:   requestInfo.Port,
	}
	SetAppConfig("", consensusOrgID, gatewayClient.ChainID)

	return e2eeClient, nil
}

func (e *E2EEClient) Encrypt(message string, targetOrgIDs []string, updateGroupKey bool) (txID string, err error) {
	url := fmt.Sprintf("https://%s:%d/v1/e2ee/org/encrypt", e.requestIP, e.requestPort)
	encMsgParam := &encMsgForSDKParams{
		TargetOrgIDs:   targetOrgIDs,
		Message:        message,
		UpToChain:      false,
		UpdateGroupKey: updateGroupKey,
	}
	encMsgParamByte, err := json.Marshal(encMsgParam)
	if err != nil {
		return "", errors.WithMessage(err, "E2EE encryption marshal encryption parameters error")
	}
	respByte, err := doRequest(url, string(encMsgParamByte), e.gatewayClient.E2EEHttpClient)
	if err != nil {
		return "", errors.WithMessage(err, "E2EE encryption send request error")
	}
	respParam := &encMsgForOrgUserResponseParams{}
	if err = json.Unmarshal(respByte, respParam); err != nil {
		return "", errors.WithMessage(err, "E2EE encryption unmarshal response parameters error")
	}
	txID, err = messageUpToChain(e.gatewayClient, respParam.DataToChain)
	if err != nil {
		return "", err
	}
	return txID, nil
}

func (e *E2EEClient) Decrypt(txID string) (message string, err error) {
	url := fmt.Sprintf("https://%s:%d/v1/e2ee/org/decrypt", e.requestIP, e.requestPort)
	decMsgParam := &decMsgForSDKRequestParams{
		TransactionID: txID,
	}
	decMsgParamByte, err := json.Marshal(decMsgParam)
	if err != nil {
		return "", errors.WithMessage(err, "E2EE decryption marshal decryption parameters error")
	}
	respByte, err := doRequest(url, string(decMsgParamByte), e.gatewayClient.E2EEHttpClient)
	if err != nil {
		return "", errors.WithMessage(err, "E2EE decryption send request error")
	}
	responseParam := &decMsgForSDKResponseParams{}
	if err = json.Unmarshal(respByte, responseParam); err != nil {
		return "", errors.WithMessage(err, "E2EE decryption unmarshal response parameters error")
	}
	return responseParam.Message, nil
}

func messageUpToChain(gatewayClient *GatewayClient, args []byte) (txID string, err error) {
	execArgs := [][]byte{args}
	_, txID, err = Exec(gatewayClient, storeMsgFuncName, execArgs)
	if err != nil {
		return "", err
	}
	return txID, nil
}

func doRequest(url, body string, httpClient *http.Client) ([]byte, error) {
	resp, err := httpClient.Post(url, "application/json", strings.NewReader(body))
	if err != nil {
		return nil, errors.WithMessage(err, "send request to E2EE error")
	}
	defer resp.Body.Close()
	resByte, err := copyRespBody(resp)
	if err != nil {
		return nil, errors.WithMessagef(err, "get E2EE response error")
	}
	if resp.StatusCode != http.StatusOK {
		return nil, errors.Errorf("E2EE response status code: %d != 200/201, response body: %s",
			resp.StatusCode, string(resByte))
	}
	return resByte, nil
}

func copyRespBody(res *http.Response) ([]byte, error) {
	if res == nil || res.Body == nil {
		return nil, nil
	}
	body, err := ioutil.ReadAll(res.Body)
	if err != nil {
		return nil, err
	}
	return body, nil
}
