package client

import (
	"encoding/hex"
	"strings"
	"time"

	"github.com/pkg/errors"

	"git.huawei.com/huaweichain/proto"
	"git.huawei.com/huaweichain/proto/common"
	"git.huawei.com/huaweichain/sdk/action/event"
	"git.huawei.com/huaweichain/sdk/config"
	"git.huawei.com/huaweichain/sdk/node"
	"git.huawei.com/huaweichain/sdk/rawmessage"
)

const (
	ContractName = "e2ee"
	WaitTime     = 60
)

type Config struct {
	ConfigFilePath string // yaml配置文件路径
	ContractName   string // 合约名称
	EndorserNodes  string // 背书节点，若背书策略为“全部组织背书”，则背书节点为每个组织的中的任一节点即可，eg: "node-0.organization,node-0.organization1"
	ConsensusNode  string // 共识节点，默认为"node-0.organization"
	ChainID        string // 链ID，默认为"default"
	QueryNode      string // 查询节点,可选择一个或者若干个组织内的节点(参见yaml配置文件中)，用于发出执行请求
	SignAlgorithm  string // 安全机制
}

// Nodes is the definition of Nodes.
type Nodes struct {
	Endorsers     []*node.WNode
	Proposer      *node.WNode
	EventListener *node.WNode
}

var gTxEvent *event.TxEventService

func getGlobeTxEvent(config *Config, net *Nodes) (*event.TxEventService, error) {
	if gTxEvent == nil {
		txEtmp, err := net.EventListener.EventAction.GetTxEventService(config.ChainID)
		if err != nil {
			return nil, errors.WithMessage(err, "event action register tx event error")
		}

		gTxEvent = txEtmp
	}
	return gTxEvent, nil
}

var appConfig Config

func AppConfig() Config {
	return appConfig
}

func SetSignAlg(cfg Config) {
	clientConfig, err := config.NewClientConfig(cfg.ConfigFilePath)
	if err != nil {
		return
	}
	appConfig.SignAlgorithm = clientConfig.Client.Type
}

func init() {
	appConfig.ContractName = ContractName
}

func SetAppConfig(filePath, consensusOrgID, chainID string) {
	appConfig.EndorserNodes = "node-1." + consensusOrgID
	appConfig.ConsensusNode = "node-0." + consensusOrgID
	appConfig.QueryNode = "node-1." + consensusOrgID
	appConfig.ConfigFilePath = filePath
	appConfig.ChainID = chainID
	SetSignAlg(appConfig)
}

func Exec(gatewayClient *GatewayClient, funcName string, args [][]byte) (txResp *common.RawMessage, txHashID string, err error) {
	argStr := bytesToFormatArgs(args)
	txResp, txHashID, err = Send(gatewayClient, AppConfig(), funcName, argStr)
	if err != nil {
		return nil, "", errors.Errorf("Send tx error, funcName [%s], err: %s", funcName, err.Error())
	}
	return txResp, txHashID, nil
}

func bytesToFormatArgs(args [][]byte) []string {
	if len(args) <= 0 {
		return nil
	}
	ret := make([]string, 0)
	for _, argByte := range args {
		ret = append(ret, string(argByte))
	}
	return ret
}

func Send(gatewayClient *GatewayClient, config Config, funcName string, argsSlice []string) (*common.RawMessage, string, error) {
	// 1.入参处理
	var err error
	endorseNodes := strings.Split(config.EndorserNodes, ",")

	// 2.背书消息请求构造
	rawMsg, err := getInvokeRawMsg(gatewayClient, config, endorseNodes, funcName, argsSlice)
	if err != nil {
		return nil, "", errors.WithMessage(err, "contract raw messessage build invoke message error")
	}

	// 3.背书消息请求发送
	invokeResponses, net, err := sendInvokeRawMsg(gatewayClient, config, endorseNodes, rawMsg)
	if err != nil {
		return nil, "", errors.WithMessage(err, "send raw messessage error")
	}

	// 4.落盘消息构建
	txRawMsg, err := getTransactionRawMsg(gatewayClient, config, invokeResponses)
	if err != nil {
		return nil, "", errors.WithMessage(err, "build transaction message error")
	}

	// 5.落盘消息发送
	responseMsg, txHashID, err := sendTransactionRawMsg(config, txRawMsg, net)
	if err != nil {
		return nil, "", errors.WithMessage(err, "send transaction message error")
	}

	// 6.解析response获得response.Payload
	payload, err := GetPayloadWithResp(responseMsg)
	if err != nil {
		return nil, "", errors.WithMessage(err, "parse response msg with struct type error")
	}

	// 7.解析payload获得common.RawMessage{}
	txResponse := &common.RawMessage{}
	if err := proto.Unmarshal(payload, txResponse); err != nil {
		return nil, "", errors.WithMessage(err, "unmarshal transaction response error")
	}
	return txResponse, txHashID, nil
}

func getInvokeRawMsg(gatewayClient *GatewayClient, config Config, endorserNodes []string, function string, args []string) (*common.RawMessage, error) {
	return gatewayClient.ContractRawMessage.BuildInvokeMessage(config.ChainID, config.ContractName, function, args)
}

func sendInvokeRawMsg(gatewayClient *GatewayClient, config Config, endorseNodes []string, rawMsg *common.RawMessage) ([]*common.RawMessage, *Nodes, error) {
	var net *Nodes
	net, err := NewNodes(gatewayClient, endorseNodes, config.ConsensusNode)
	if err != nil {
		return nil, nil, errors.WithMessage(err, "new nodes network error")
	}

	// 1.背书请求消息发送
	var invokeResponses []*common.RawMessage
	for _, node := range net.Endorsers {
		var invokeResponse *common.RawMessage
		invokeResponse, err = node.ContractAction.Invoke(rawMsg)
		if err != nil {
			return nil, nil, errors.WithMessage(err, "invoke error")
		}
		invokeResponses = append(invokeResponses, invokeResponse)
	}
	return invokeResponses, net, nil
}

func sendTransactionRawMsg(config Config, txRawMsg *rawmessage.TxRawMsg, net *Nodes) (*common.RawMessage, string, error) {
	txEvent, err := getGlobeTxEvent(&config, net)
	if err != nil {
		return nil, "", errors.WithMessage(err, "event action register tx event error")
	}

	resultChan, err := txEvent.RegisterTx(txRawMsg.Hash)
	if err != nil {
		return nil, "", errors.WithMessage(err, "tx event register tx id error")
	}

	transactionResponse, err := net.Proposer.ContractAction.Transaction(txRawMsg.Msg)
	if err != nil {
		return nil, "", errors.WithMessage(err, "invoke error")
	}

	select {
	case txResult := <-resultChan:
		if txResult.Status != common.VALID {
			return transactionResponse, Hash2str(txResult.TxHash), errors.New("MVCC")
		}
		return transactionResponse, Hash2str(txResult.TxHash), nil
	case <-time.After(WaitTime * time.Second):
		return nil, "", errors.Errorf("send transaction time out")
	}
}

func NewNodes(client *GatewayClient, nodeNames []string, consNode string) (*Nodes, error) {
	nodes := &Nodes{}
	nodeMap := client.Nodes
	nodes.Endorsers = make([]*node.WNode, len(nodeNames))
	for i, nodeName := range nodeNames {
		node, ok := nodeMap[nodeName]
		if !ok {
			return nil, errors.Errorf("node not exist： %v", nodeName)
		}
		nodes.Endorsers[i] = node
	}
	var ok bool
	nodes.Proposer, ok = nodeMap[consNode]
	if !ok {
		return nil, errors.Errorf("node not exist： %v", consNode)
	}
	nodes.EventListener = nodes.Endorsers[0]
	return nodes, nil
}

func getTransactionRawMsg(client *GatewayClient, config Config, transactionRawMsg []*common.RawMessage) (*rawmessage.TxRawMsg, error) {
	return client.ContractRawMessage.BuildTxRawMsg(transactionRawMsg)
}

func GetPayloadWithResp(responseMsg *common.RawMessage) ([]byte, error) {
	response := &common.Response{}
	if err := proto.Unmarshal(responseMsg.Payload, response); err != nil {
		return nil, errors.WithMessage(err, "execChainAction Unmarshal message error")
	}
	if response.Status != common.SUCCESS {
		return nil, errors.Errorf("execChainAction response.Status != common.SUCCESS, status: %s, info: %v",
			response.Status.String(), response.StatusInfo)
	}
	return response.Payload, nil
}

// Hash2str is used to convert the hash value to string for human reading.
func Hash2str(h []byte) string {
	return hex.EncodeToString(h)
}
